function [rsvmmodel hloss lsloss accs] = ReduceSVs(y, Ksv, svmmodel, params, testmodel)
%
% Reduce suppor vectors by least-squares method
%   [rsvmmodel hloss lsloss accs] = ReduceSVs(y, K, svmmodel,  params, testmodel)
%
% Input:
%  y        - ID labels (+1/-1) for training Samples [num x 1]
%  K        - Gram matrix of training samples vs support vectors [num x nsv]
%  svmmodel - SVM model [struct]
%             .SVs     - Support vectors [nsv x dim], or indexes [nsv x 1]
%             .sv_coef - Coefficient for the support vectors [nsv x 1]
%             .rho     - Bias [scalar]
%             .K       - Gram matrix of support vectors [nsv x nsv]
% (optional)
%  params   - Parameters [struct]
%             .tol - Termination tolerance in increase of Hinge loss [{0.025}]
%             .eps - Regularization parameter for inverse matrix [{0.001}]
%             .maxiter - Maximun number of iterations [{inf}]
%  testmodel- Test samples and labels [struct]
%             .y - ID labels (+1/-1) for the test samples [testnum x 1]
%             .K - Gram matrix of test samples vs support vectors [testnum x nsv]
%
% Output:
%  rsvmmodel - Reduced SVM model [structure]
%             .SVs     - Support vectors [nsv x dim], or indexes [nsv x 1]
%             .sv_coef - Coefficient for the support vectors [nsv x 1]
%             .rho     - Bias [scalar]
%  hloss     - Hinge losses through the reduction process [iter x 1]
%  lsloss    - Least-square errors through the reduction process [iter x 1]
%  accs      - Performance accuracies on test samples if provided [iter x 1]
%
% Reference:
% [1] T. Kobayashi and N. Otsu, 
%     "EFFICIENT REDUCTION OF SUPPORT VECTORS IN KERNEL-BASED METHODS", ICIP 2009.
%

% Copyright (c) 2009-2012, Takumi Kobayashi
% All rights reserved.
% 
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
% 
%     * Redistributions of source code must retain the above copyright 
%       notice, this list of conditions and the following disclaimer.
%     * Redistributions in binary form must reproduce the above copyright
%       notice, this list of conditions and the following disclaimer in the
%       documentation and/or other materials provided with the distribution
%       
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
% IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
% THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
% EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
% PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
% PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
% LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
% NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
% SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

%- Parameters -%
if ~exist('params', 'var') || isempty(params), params = struct; end
params = parseparam(params, 'tol',0.025, 'eps',0.001, 'maxiter',Inf);

%- Init -%
nsv     = length(svmmodel.sv_coef);
maxiter = min(nsv-1,params.maxiter);
hloss   = zeros(maxiter,1);
lsloss  = zeros(maxiter,1);

classifier = @(A, model) A*model.sv_coef - model.rho;
hloss(1)   = mean(max(1 - y.*classifier(Ksv,svmmodel),0)); %hinge loss
lsloss(1)  = 0; %least-square loss

if exist('testmodel', 'var') && ~isempty(testmodel)
    %- If test data are provided -%
	testK = testmodel.K;
	testy = testmodel.y;
    accs  = zeros(maxiter,1);
	accs(1) = sum(classifier(testK,svmmodel).*testy >= 0)/length(testy);
else
	accs = [];
end

H = inv(svmmodel.K+speye(nsv)*params.eps);
rsvmmodel = struct('SVs', svmmodel.SVs, 'sv_coef', svmmodel.sv_coef, 'rho', svmmodel.rho);

%- Efficient Reduction -%
rsvmmodel_old = rsvmmodel;
for j = 2:maxiter
	%-- Minimum error in least squares --%
	[mval i]   = max(diag(H));
	lsloss(j)  = 1/mval - params.eps;
	remain_idx = (1:size(H,1))~=i;
    
	%-- Remove the sample of minimum least-square error --% 
%	disp([num2str(i) '-th SV is removed']);
	b = H(remain_idx,i); 
    b_= b/H(i,i);
	
	H = H(remain_idx,remain_idx) - b*b_';
	rsvmmodel.sv_coef  = rsvmmodel.sv_coef(remain_idx) - rsvmmodel.sv_coef(i)*b_;
	Ksv(:,i) = [];
	rsvmmodel.SVs(i,:) = [];
    
	hloss(j) = mean(max(1 - y.*classifier(Ksv,rsvmmodel),0));

	%-- Recognition rate if test samples are provided --%
	if exist('testmodel', 'var') && ~isempty(testmodel)
		testK(:,i) = [];
    	accs(j) = sum(classifier(testK,rsvmmodel).*testy >= 0)/length(testy);
	end

	%- Check termination -%
	if abs(hloss(j) - hloss(1)) > params.tol
		rsvmmodel = rsvmmodel_old;
		hloss(j:end)  = [];
		lsloss(j:end) = [];
		accs(j:end)   = [];
		break;
	else
		rsvmmodel_old = rsvmmodel;
	end
end
